@@ -3,8 +3,8 @@ import warnings
 import numpy as np
 
 from . import dispatch, B, Numeric
-from ..types import NPDType, Int
 from ..shape import unwrap_dimension
+from ..types import NPDType, Int
 
 __all__ = []
 
@@ -14,20 +14,20 @@ def _warn_dtype(dtype):
         warnings.warn("Casting random number of type float to type integer.")
 
 
-@dispatch(NPDType, [Int])
-def rand(dtype, *shape):
+@dispatch
+def rand(dtype: NPDType, *shape: Int):
     _warn_dtype(dtype)
     return B.cast(dtype, np.random.rand(*shape))
 
 
-@dispatch(NPDType, [Int])
-def randn(dtype, *shape):
+@dispatch
+def randn(dtype: NPDType, *shape: Int):
     _warn_dtype(dtype)
     return B.cast(dtype, np.random.randn(*shape))
 
 
-@dispatch(Numeric, Int)
-def choice(a, n):
+@dispatch
+def choice(a: Numeric, n: Int):
     inds = np.random.choice(unwrap_dimension(B.shape(a)[0]), n, replace=True)
     choices = a[inds]
     return choices[0] if n == 1 else choices
